/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */



#include <string.h>
#include <stdio.h>

#include "xacml_client.h"
#include "xacml_server.h"

#include "xacml_io_ssl.h"

#include "scas_log.h"

/* Network and SSL */
#include "net_common.h"
#include "ssl-common.h"
/* Network and SSL */

#include <fcntl.h>


#define SOCKET_CAN_CLOSE  0
#define SOCKET_DONT_CLOSE 1
#define TRUE 1
#define FALSE 0


/* State of this thread */
typedef struct
{
    int             socket;
    SSL            *ssl;
    int             close_sock;
    int             failed;
}
ssl_io_state_t;


static
int
ssl_io_close(
             void  *arg);



#define NEW_METHOD_FOR_IPV6
#ifdef  NEW_METHOD_FOR_IPV6
static
void *
ssl_io_accept(
    int                                 msocket,
    struct sockaddr_storage            *addr,
    socklen_t                          *addr_len,
    int                                *sock_out)
{
    int                 rc             = 0;
    int                 len            = 0;
    ssl_io_state_t *    state          = NULL;
    char *              remote_host    = NULL;
    char *              ssl_subject_dn = NULL;
    SSL_CTX *           my_ctx         = NULL;


    /* Set time indicator to now! */
    scas_log_set_time_indicator();


    scas_log_debug (5, "Entering %s\n", __func__);


    /* Allocate thread specific state */
    state = malloc (sizeof (ssl_io_state_t));
    if (!state)
    {
        scas_log (1, "%s: memory allocation of ssl_io_state_t failed\n", __func__);
        rc = -1;
        return NULL;
    }
    else
    {
        bzero(state, sizeof (ssl_io_state_t));

        /* By default, close the socket. Don't close it if the SOAP interface needs to do it... this conflict triggers a seg. fault */
        state->close_sock = SOCKET_CAN_CLOSE;
        state->failed = FALSE;
        /*
        state->socket = 0;
        state->ssl    = NULL;
        state->close_sock = SOCKET_CAN_CLOSE;
        state->failed = FALSE;
        */
    }



    /******* SSL Init *********/
    my_ctx = ssl_ctx;
    if (!my_ctx)
    {
        my_ctx = setup_server_ctx(NULL, certCAPath, certFilePub, certFilePriv);
        if (!my_ctx)
        {
            scas_log (1, "%s: Error: unable to create an SSL_CTX context with setup_server_ctx()\n", __func__);
            goto skip_everything_and_exit;
        }
        else
        {
            scas_log_debug (5, "%s: Created CTX here\n", __func__); 
        } 
    }
    else
    {
        scas_log_debug (5, "%s: Got SSL_CTX object\n", __func__); 
    } 
    /**************************/


    /* Setup of Master Socket to listen on the port */
    if (msocket <= 0)
    {
        scas_log (1, "%s: Failure in master listening socket\n", __func__);
        goto skip_everything_and_exit;
    }



    /* Accept and create client socket */
    if ( (accept_new_client_socket (msocket, 
                                    &(state -> socket), 
                                    (struct sockaddr_storage *) addr,
                                    (socklen_t *) addr_len,
                                    &remote_host) ) || 
                                                    (state -> socket < 0) )
    {
        /* Don't overdue the error message, the error is mentioned in more detail in accept_new_client_socket() */
        /* scas_log (1, "%s: Failed to accept the client socket connection!\n", __func__); */
        rc = 14;
        goto finalize_TCPIP;
    }
    else
    {
        scas_log (1, "%s: TCP connection accepted from host: %s\n", __func__, remote_host);

        /* SSL stuff */
        if ((state -> ssl = SSL_server_accept (state -> socket,
                                               remote_host,
                                               my_ctx
                                              )) == NULL)
        {
            scas_log (1, "%s: Failure in SSL layer setup and connection for host: %s\n", __func__, remote_host);
            rc = 15;
            goto finalize_TCPIP;
        }
        else
        {
            scas_log_debug (3, "%s: SSL layer accepted from host %s\n", __func__, remote_host);

#define USE_LCAS_ON_SSL_LAYER
#ifdef USE_LCAS_ON_SSL_LAYER
            /* Initialize LCAS - Authorizing SSL client certificate */
            if (lcas_init_and_logfile(NULL, scas_get_log_file_pointer(), scas_get_log_type()))
            {
                /* FAILED init is failed init! 
                if (lcas_init_and_logfile(getenv("LCAS_LOG_FILE"), NULL, DO_USRLOG))
                {
                    scas_log (1, "%s: Initialization of LCAS failed!\n", __func__);
                    goto finalize_SSL;
                }
                */
                scas_log (1, "%s: Initialization of LCAS failed!\n", __func__);
                goto finalize_SSL;
            }
            else
            {
                /* TODO: I should make a beter and cleaner interface in LCAS to do this */
                BIO * certbio_mem = NULL;
                STACK_OF (X509) * px509_chain = NULL;
                X509 *            px509       = NULL;
                char * pem_string = NULL;
                int i = 0;

                scas_log_debug (5, "LCAS logs successfully integrated with SCAS log file\n");
                certbio_mem = BIO_new(BIO_s_mem());
             
                /* Get certificate chain */
                if (!(px509_chain = SSL_get_peer_cert_chain(state -> ssl)))
                {
                    scas_log (1, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", __func__);
                    goto finalize_SSL;
                }

                /* Push certificates in chain into the BIO memory stack */
                for (i = 0; i < sk_X509_num(px509_chain); i++) 
                {
                    int retval = 0;
                    char tmp_dn[256]; /* Note... the 256 buffer length is only used for logging output on the spot here. More authoritive information is in the LCAS and LCMAPS logs */

                    px509 = sk_X509_value(px509_chain, i);
                    X509_NAME_oneline(X509_get_subject_name(px509), tmp_dn, 256);
                    scas_log_debug (5, "%s: Subject is: %s\n", __func__, tmp_dn);
                    retval = PEM_write_bio_X509(certbio_mem, px509);
                    scas_log_debug (5, "%s: PEM_write_bio_X509 from stack item %d, %s\n", __func__, i, (retval ? "succeeded" : "failed"));
                }
                
                /* Report the subject DN of the connecting party */
                ssl_subject_dn = x509_chain_to_dn (px509_chain);
                scas_log (1, "%s: \"%s\" is connected from \"%s\"\n", __func__, ssl_subject_dn, remote_host);
                free (ssl_subject_dn);
                ssl_subject_dn = NULL;

                /* Push BIO memory to character array */
                /* Oscar
                 * Warning: The BIO_get_mem_data() will only apply a REFERENCE to 
                 * the pem_string, NOT A COPY of the data 
                 */
                len = BIO_get_mem_data (certbio_mem, &pem_string); 
                pem_string[len]='\0';

                scas_log_debug (5, "%s: PEM string constructed for LCAS use on SSL layer.\n", __func__);

                if (lcas_pem(pem_string, "RSL") != 0)
                {
                    scas_log (1, "%s: Client credentials are not authorized!\n", __func__);
                    if (lcas_term())
                    {
                        scas_log (2, "%s: LCAS termination failure.\n", __func__);
                    }

                    /* 
                     * Closing and Freeing buffers AFTER making use of them in LCAS
                     */
                    BIO_set_close(certbio_mem, BIO_CLOSE);
                    BIO_free(certbio_mem);

                    goto finalize_SSL;
                }
                else
                {
                    scas_log_debug (3, "%s: Client credentials authorized to send query\n", __func__);

                    /* LCAS term */
                    if (lcas_term())
                    {
                        scas_log (2, "%s: LCAS termination failure.\n", __func__);
                        rc = 103;
                    }

                    /* 
                     * Closing and Freeing buffers AFTER making use of them in LCAS
                     */
                    BIO_set_close(certbio_mem, BIO_CLOSE);
                    BIO_free(certbio_mem);
                }
            }
#endif /* USE_LCAS_ON_SSL_LAYER */


            /*** Good exit ***/

            /* Return the connected client socket to the XACML library to listen for traffic */
            free(remote_host);
            *sock_out = state -> socket;
            scas_log_debug (5, "%s: done\n", __func__);
            return state;
        }
        /* SSL stuff */
    }



/*** Bad exit ***/
finalize_SSL:
#if 0
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (SSL_connect_shutdown (state -> ssl))
        {
            scas_log (1, "%s: Error in Server SSL_connect_shutdown!\n", __func__);
        }
        SSL_free(state -> ssl);
    }
#endif

finalize_TCPIP:
#if 0
    if (state)
    {
        close (state -> socket);
        state -> socket = 0;
    }
    free(addr);
#endif

skip_everything_and_exit:
    /*
    free(state);
    state = NULL;
     */
    free(remote_host);

    state->failed = TRUE;

    /* Don't Close the socket yet! This causes a segmentation fault in the stdsoap2_cpp.cpp */
    state->close_sock = SOCKET_DONT_CLOSE;
    ssl_io_close (state);

    state->close_sock = SOCKET_CAN_CLOSE;
    scas_log_debug (2, "%s: ended in a failure\n", __func__);

    /* return NULL; */
    return state;
}
#else
static
void *
ssl_io_accept(
    int                                 msocket,
    struct sockaddr                    *addr,
    socklen_t                          *addr_len,
    int                                *sock_out)
{
    int                 rc             = 0;
    int                 len            = 0;
    ssl_io_state_t *    state          = NULL;
    char *              remote_host    = NULL;
    char *              ssl_subject_dn = NULL;
    SSL_CTX *           my_ctx         = NULL;


    /* Set time indicator to now! */
    scas_log_set_time_indicator();


    scas_log_debug (5, "Entering %s\n", __func__);


/*
 *   scas_log_debug (5, "Master Socket is: %d.\n", msocket);
 *   if (fcntl (msocket, F_SETFL, O_NONBLOCK) == -1)
 *   {
 *       scas_log_debug (5, "Setting master socket to non-blocking failed\n", __func__);
 *       return NULL;
 *   }
 */


    /* Allocate thread specific state */
    state = malloc (sizeof (ssl_io_state_t));
    if (!state)
    {
        scas_log (1, "%s: memory allocation of ssl_io_state_t failed\n", __func__);
        rc = -1;
        return NULL;
    }
    else
    {
        bzero(state, sizeof (ssl_io_state_t));

        /* By default, close the socket. Don't close it if the SOAP interface needs to do it... this conflict triggers a seg. fault */
        state->close_sock = SOCKET_CAN_CLOSE;
        state->failed = FALSE;
        /*
        state->socket = 0;
        state->ssl    = NULL;
        state->close_sock = SOCKET_CAN_CLOSE;
        state->failed = FALSE;
        */
    }



    /******* SSL Init *********/
    my_ctx = ssl_ctx;
    if (!my_ctx)
    {
        my_ctx = setup_server_ctx(NULL, certCAPath, certFilePub, certFilePriv);
        if (!my_ctx)
        {
            scas_log (1, "%s: Error: unable to create an SSL_CTX context with setup_server_ctx()\n", __func__);
            goto skip_everything_and_exit;
        }
        else
        {
            scas_log_debug (5, "%s: Created CTX here\n", __func__); 
        } 
    }
    else
    {
        scas_log_debug (5, "%s: Got SSL_CTX object\n", __func__); 
    } 
    /**************************/


    /* Setup of Master Socket to listen on the port */
    if (msocket <= 0)
    {
        scas_log (1, "%s: Failure in master listening socket\n", __func__);
        goto skip_everything_and_exit;
    }


    *addr_len = sizeof(struct sockaddr_in);


    /* Accept and create client socket */
    if ( (accept_new_client_socket (msocket, 
                                    &(state -> socket), 
                                     (struct sockaddr_in *) addr, 
                                     (socklen_t *) addr_len, 
                                    &remote_host) ) || 
                                                    (state -> socket < 0) )
    {
        /* Don't overdue the error message, the error is mentioned in more detail in accept_new_client_socket() */
        /* scas_log (1, "%s: Failed to accept the client socket connection!\n", __func__); */
        rc = 14;
        goto finalize_TCPIP;
    }
    else
    {
        scas_log (1, "%s: TCP connection accepted from host: %s\n", __func__, remote_host);
        /* SSL stuff */
        if ((state -> ssl = SSL_server_accept (state -> socket,
                                               remote_host,
                                               my_ctx
                                              )) == NULL)
        {
            scas_log (1, "%s: Failure in SSL layer setup and connection for host: %s\n", __func__, remote_host);
            rc = 15;
            goto finalize_TCPIP;
        }
        else
        {
            scas_log_debug (3, "%s: SSL layer accepted from host %s\n", __func__, remote_host);

#define USE_LCAS_ON_SSL_LAYER
#ifdef USE_LCAS_ON_SSL_LAYER
            /* Initialize LCAS - Authorizing SSL client certificate */
            if (lcas_init_and_logfile(NULL, scas_get_log_file_pointer(), scas_get_log_type()))
            {
                /* FAILED init is failed init! 
                if (lcas_init_and_logfile(getenv("LCAS_LOG_FILE"), NULL, DO_USRLOG))
                {
                    scas_log (1, "%s: Initialization of LCAS failed!\n", __func__);
                    goto finalize_SSL;
                }
                */
                scas_log (1, "%s: Initialization of LCAS failed!\n", __func__);
                goto finalize_SSL;
            }
            else
            {
                /* TODO: I should make a beter and cleaner interface in LCAS to do this */
                BIO * certbio_mem = NULL;
                STACK_OF (X509) * px509_chain = NULL;
                X509 *            px509       = NULL;
                char * pem_string = NULL;
                int i = 0;

                scas_log_debug (5, "LCAS logs successfully integrated with SCAS log file\n");
                certbio_mem = BIO_new(BIO_s_mem());
             
                /* Get certificate chain */
                if (!(px509_chain = SSL_get_peer_cert_chain(state -> ssl)))
                {
                    scas_log (1, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", __func__);
                    goto finalize_SSL;
                }

                /* Push certificates in chain into the BIO memory stack */
                for (i = 0; i < sk_X509_num(px509_chain); i++) 
                {
                    int retval = 0;
                    char tmp_dn[256]; /* Note... the 256 buffer length is only used for logging output on the spot here. More authoritive information is in the LCAS and LCMAPS logs */

                    px509 = sk_X509_value(px509_chain, i);
                    X509_NAME_oneline(X509_get_subject_name(px509), tmp_dn, 256);
                    scas_log_debug (5, "%s: Subject is: %s\n", __func__, tmp_dn);
                    retval = PEM_write_bio_X509(certbio_mem, px509);
                    scas_log_debug (5, "%s: PEM_write_bio_X509 from stack item %d, %s\n", __func__, i, (retval ? "succeeded" : "failed"));
                }
                
                /* Report the subject DN of the connecting party */
                ssl_subject_dn = x509_chain_to_dn (px509_chain);
                scas_log (1, "%s: \"%s\" is connected from \"%s\"\n", __func__, ssl_subject_dn, remote_host);
                free (ssl_subject_dn);
                ssl_subject_dn = NULL;

                /* Push BIO memory to character array */
                /* Oscar
                 * Warning: The BIO_get_mem_data() will only apply a REFERENCE to 
                 * the pem_string, NOT A COPY of the data 
                 */
                len = BIO_get_mem_data (certbio_mem, &pem_string); 
                pem_string[len]='\0';

                scas_log_debug (5, "%s: PEM string constructed for LCAS use on SSL layer.\n", __func__);

                if (lcas_pem(pem_string, "RSL") != 0)
                {
                    scas_log (1, "%s: Client credentials are not authorized!\n", __func__);
                    if (lcas_term())
                    {
                        scas_log (2, "%s: LCAS termination failure.\n", __func__);
                    }

                    /* 
                     * Closing and Freeing buffers AFTER making use of them in LCAS
                     */
                    BIO_set_close(certbio_mem, BIO_CLOSE);
                    BIO_free(certbio_mem);

                    goto finalize_SSL;
                }
                else
                {
                    scas_log_debug (3, "%s: Client credentials authorized to send query\n", __func__);

                    /* LCAS term */
                    if (lcas_term())
                    {
                        scas_log (2, "%s: LCAS termination failure.\n", __func__);
                        rc = 103;
                    }

                    /* 
                     * Closing and Freeing buffers AFTER making use of them in LCAS
                     */
                    BIO_set_close(certbio_mem, BIO_CLOSE);
                    BIO_free(certbio_mem);
                }
            }
#endif /* USE_LCAS_ON_SSL_LAYER */


            /*** Good exit ***/

            /* Return the connected client socket to the XACML library to listen for traffic */
            free(remote_host);
            *sock_out = state -> socket;
            scas_log_debug (5, "%s: done\n", __func__);
            return state;
        }
        /* SSL stuff */
    }



/*** Bad exit ***/
finalize_SSL:
#if 0
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (SSL_connect_shutdown (state -> ssl))
        {
            scas_log (1, "%s: Error in Server SSL_connect_shutdown!\n", __func__);
        }
        SSL_free(state -> ssl);
    }
#endif

finalize_TCPIP:
#if 0
    if (state)
    {
        close (state -> socket);
        state -> socket = 0;
    }
    free(addr);
#endif

skip_everything_and_exit:
    /*
    free(state);
    state = NULL;
     */
    free(remote_host);

    state->failed = TRUE;

    /* Don't Close the socket yet! This causes a segmentation fault in the stdsoap2_cpp.cpp */
    state->close_sock = SOCKET_DONT_CLOSE;
    ssl_io_close (state);

    state->close_sock = SOCKET_CAN_CLOSE;
    scas_log_debug (2, "%s: ended in a failure\n", __func__);

    /* return NULL; */
    return state;
}
#endif /* NEW_METHOD_FOR_IPV6 */




static 
void *
ssl_io_connect(
    const char                         *endpoint,
    const char                         *host,
    int                                 port)
{
    ssl_io_state_t  *state = NULL;

    int              realtime_timeout      = 5; /* seconds */
    int              backoff_micro_seconds = 500; /* micro seconds */
    int              socket_timeout        = 200; /* milli seconds */

    int              err     = 0;
    int              rc      = 0;
    time_t           start   = 0;


    /* Set time indicator to now! */
    scas_log_set_time_indicator();


    scas_log_debug (5, "Entering %s\n", __func__);

    /* Allocate thread specific state */
    state = malloc (sizeof (ssl_io_state_t));
    if (!state)
    {
        scas_log (1, "%s: memory allocation of ssl_io_state_t failed\n", __func__);
        rc = -1;
        return NULL;
    }
    else
    {
        state->socket = 0;
        state->ssl    = 0;
    }


    start = time(NULL);
    while ((time(NULL) - start) < realtime_timeout)
    {
        /* Create client socket */
        scas_log_debug (5, "%s: Connecting socket to host %s on port %d\n", __func__, host, port);
        if (create_client_socket (&(state -> socket),
                                   host,
                                   port,
                                   socket_timeout))
        {
            /* Retry connection setup, with a random wait */
            scas_log_debug (1, "%s: Socket creation failed, retrying... in %d milli seconds!\n", __func__, backoff_micro_seconds / 1000);

            /* Exponential back off in microseconds */
            usleep (backoff_micro_seconds);
            backoff_micro_seconds *= 2;
            continue; /* Continue until the timeout is reached */
        }
        else
        {
            /* TCP/IP fully connected */
            scas_log_debug (5, "%s: Connected on TCP/IP\n", __func__);
            scas_log (3, "%s: Is connected to host '%s' on port '%d'\n", __func__, host, port);

            /* SSL */
            if ((state -> ssl = SSL_client_connect (state -> socket,
                                                    setup_client_ctx (NULL,
                                                                      certCAPath,
                                                                      certFilePub,
                                                                      certFilePriv)
                                          )) == NULL)
            {
                scas_log (1, "%s: Failure in SSL layer setup and connection!\n", __func__);

                /* Finalize TCP/IP */
                rc = 15;
                goto finalize_TCPIP;
            }
            else
            {
                scas_log_debug (5, "%s: Connected on SSL!\n", __func__);

                /* Post SSL connection check */
                if ((err = post_connection_check(state -> ssl, host)) != X509_V_OK)
                {
                    scas_log (2, "%s: Error with peer certificate after the full handshake: %s\n", __func__, X509_verify_cert_error_string(err));
                    SSL_set_verify_result (state -> ssl, err); 

                    /* Finalize SSL connection */
                    rc = 16;
                    goto finalize_SSL;
                }
                else
                {
                    scas_log_debug (5, "%s: SSL layer (client side) connection done\n", __func__);

                    /* Finalize SSL connection */
                    rc = 0;
                    goto succesful_connect;
                }
            }
            /* SSL */
        }
    }


    /* Failure */
    scas_log (1, "%s: Failed to connect to host '%s' on port '%d'\n", __func__, host, port);
    goto finalize_TCPIP;


/* Good exit */
succesful_connect:
    scas_log_debug (5, "%s: done\n", __func__);
    return state;


/* Bad exit */
finalize_SSL:

    /* There should be an SSL object to destroy in the first place */
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (SSL_connect_shutdown (state -> ssl))
        {
            scas_log (1, "%s: Error: in client SSL_connect_shutdown!\n", __func__);
        }
        SSL_free(state -> ssl);
    }

finalize_TCPIP:
    if (state)
    {
        close (state -> socket);
        free(state);
        state = NULL;
    }

    scas_log (3, "%s: ended in a failure\n", __func__);
    return NULL;
}
/* ssl_io_connect() */


static
int
ssl_io_send(
    void                               *arg,
    const char                         *data,
    size_t                              size)
{
    ssl_io_state_t *                    state = arg;
    int                                 rc = -2;
    int                                 nwritten = 0;

    scas_log_debug (5, "%s: Sending...\n", __func__);

    if (state && state -> ssl)
    {
        for (nwritten = 0; nwritten < size; nwritten += rc)
        {
            rc = SSL_write (state -> ssl, data + nwritten, size - nwritten);
            /* Error conditions are either zero or < zero. */
            if (rc <= 0)
            {
                print_ssl_error_msg (state -> ssl, rc, 1, NULL);
                return -1;
            }
        }
    }


    BIO_flush(state -> ssl -> wbio);

    scas_log_debug (5, "%s: Send %d of %d\n", __func__, rc, size);  
    return 0;
}
/* ssl_io_send() */


static
size_t
ssl_io_recv(
    void                               *arg,
    char                               *data,
    size_t                              size)
{
    ssl_io_state_t * state = arg;
    int              rc = 0;


    if (state && state -> ssl)
    {
        if ((rc = SSL_recv (state -> ssl, data, size)) < 0) 
        {
            BIO_flush(state -> ssl -> rbio);
            scas_log (2, "%s: Failure in SSL_recv\n", __func__);
            return -1; 
        }
    }

    scas_log_debug (5, "%s: End of ssl_io_recv()\n", __func__);
    return rc;
}
/* ssl_io_recv() */



static
int
ssl_io_close(
             void  *arg)
{
    ssl_io_state_t *                state = arg;
    int                             failed_state = FALSE;

    scas_log_debug (5, "%s: Entering ssl_io_close()\n", __func__);

    /* Only work with it, if not already destroyed or set to NULL */
    if (state)
    {
        failed_state = state -> failed;

        if (state -> ssl)
        {    
            /* Shutdown active SSL layer */
            if (SSL_connect_shutdown (state -> ssl))
            {
                scas_log (3, "%s: Error: in SSL_connect_shutdown!\n", __func__);
            }
            state -> ssl -> rbio = NULL;
            state -> ssl -> wbio = NULL;
            scas_log_debug (3, "%s: going to SSL_free()\n", __func__);
            
            SSL_free(state -> ssl);
            state -> ssl = NULL;
        }

        if (state -> socket)
        {
            if (state -> close_sock == SOCKET_CAN_CLOSE)
            {
                close (state -> socket);
                state -> socket = 0;
        
                free(state);
                state = NULL;
            }
        }

/*
        if (state -> close_sock == SOCKET_CAN_CLOSE)
        {
            free(state);
            state = NULL;
        }
*/
    }


    scas_log_debug (5, "%s: ended perfectly.\n", __func__);
    return failed_state == TRUE ? 1 : 0;
}
/* ssl_io_close() */


xacml_io_descriptor_t
xacml_io_ssl_descriptor =
{
    "xacml_io_ssl_descriptor",
    ssl_io_accept,
    ssl_io_connect,
    ssl_io_send,
    ssl_io_recv,
    ssl_io_close
};

