/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */

#include "open_conf_file.h"



/********************************************************** 
Function:       trimmer
Description:
                This function will trim non graphical characters
                from the string and return a new string.
                The string must be null-terminated.
Parameters:
 char * str  :  The to be trimmed string.
                Note: If the  string doesn't contain graphically displayable characters, then the string is trimmed to a NULL.

Return value:
         !NULL  : string for which its allocated memory must be free()'d
          NULL  : bad
**********************************************************/
char * trimmer (char * str)
{
    char * trimmed = NULL;
    int i = 0, trimmer_len = 0, first_alphanum = -1, last_alphanum  = -1;
    int len = strlen(str);

    /* Scan and locate first alphanum character */
    for (i = 0; i < len; i++)
    {
        /* IS alpha num */
        if (isgraph(str[i]))
        {
            first_alphanum = i;
            break;
        }
    }

    /* If no alphanum found, bail out. There is all to trim, which is useless. */
    if (first_alphanum == -1)
    {
        return NULL;
    }

    /* Scan and locate last alphanum character */
    for (i = len - 1; i >= 0; i--)
    {
        /* IS alpha num */
        if (isgraph(str[i]))
        {
            last_alphanum = i;
            break;
        }
    }

    /* Allocate sufficient storage space in memory. On fail, bail! */
    trimmer_len = last_alphanum - (first_alphanum - 1);
    trimmed     = malloc (sizeof (char) * trimmer_len + 1);
    if (trimmed == NULL)
        return NULL;

    /* Copy memory chunk */
    memcpy (trimmed, &str[first_alphanum], trimmer_len);
    trimmed [trimmer_len] = '\0';

    return trimmed;
}



/********************************************************************** 
Function:              parse_scas_config_file
Description:           This function will parse the input buffer
                       and will return a config_name_values structure
                       The parsing of the buffer will be per line. Each line will be split on the '=' (equal-sign).
                       The split elements are the name and value. Each of these split strings will be trimmed at the beginning and the end of the string of their non-alpha-numerical characters.
Parameters:
    unsigned char * buffer       : 8-bit buffer string to be parsed.
    int             bufsize      : size of buffer.
    config_name_values * cnv     : pointer to config_name_values structure which will hold the name/value pairs    
    int n_config_name_values     : amount of name/value pairs in cnv
Return value:
           0  : good
          !0  : bad
**********************************************************************/
int parse_scas_config_file (unsigned char * buffer, int bufsize, config_name_values * cnv, int * n_config_name_values)
{
    int i = 0, j = 0;
    char * line_buf;

    line_buf = malloc (sizeof (char) * 1024);
    if (!line_buf)
    {
        return 1;
    }

    *cnv = NULL;

    for (i = 0; i < bufsize; i++, j++)
    {
        if (buffer[i] == '\n')
        {
            char * tmp = NULL;
            char * name = NULL;
            char * value = NULL;

            line_buf[j] = '\0';

            tmp = line_buf;

            name = strsep(&tmp, "=");
            if ((name) && (strlen(name) > 0))
            {
                 /* Allocate config_name_values array */
                 *cnv = realloc (*cnv, sizeof (struct s_config_name_values) * ++(*n_config_name_values));

                 (*cnv)[(*n_config_name_values) - 1].name  = trimmer(name);

                 value = strsep(&tmp, "=");
                 if ((value) && (strlen(value) > 0))
                 {
                     (*cnv)[(*n_config_name_values) - 1].value = trimmer(value);
                 }
            }

            j = -1;
        }
        else
        {
            line_buf[j] = buffer[i];
        } 
    } 


    free (line_buf);
    return 0;
}




/**************************************************************************************************
function :                     setup_scas_config_name_value
description :
                   This function will setup all the SCAS, LCAS, and LCMAPS environment variables
      SCAS_PORT:              TCP portnumber for the SCAS service to listen on
      SCAS_LOG_FILE:          SCAS log file
      SCAS_HOSTCERT:          Host certificate of the SCAS service
      SCAS_HOSTKEY:           Private key file that's with the host certificate of the SCAS service
      SCAS_CAPATH:            CA certificate directory that contains the CRL files and the CA root certificates
      SCAS_LOG_LEVEL:         Log level (0-5)
      SCAS_DEBUG_LEVEL:       Debug level (0-5)
      SCAS_NO_DAEMONIZE:      Do not demonize the SCAS process 
      SCAS_REFRESH_RATE:      Refresh rate of the SCAS child processes

      LCAS_LOG_FILE:          Name and location of the logfile
      LCAS_LOG_TYPE:          Type of logging (logfile, syslog, both or none)
      LCAS_LOG_STRING:        Log string to be prepended to the logging lines
      LCAS_LOG_LEVEL:         Log level (0-5)
      LCAS_DB_FILE:           Name of the lcas policy file
      LCAS_DEBUG_LEVEL:       Debug level (0-5)
      LCAS_MOD_DIR:           Location of the LCAS library
      LCAS_DIR:               Location of the LCAS config files
      LCAS_ETC_DIR:           Location of the LCAS config files (=LCAS_DIR)

      LCMAPS_LOG_FILE:          Name and location of the logfile
      LCMAPS_LOG_TYPE:          Type of logging (logfile, syslog, both or none)
      LCMAPS_LOG_STRING:        Log string to be prepended to the logging lines
      LCMAPS_DB_FILE:           Name of the lcmaps policy file
      LCMAPS_LOG_LEVEL:         Log level (0-5)
      LCMAPS_DEBUG_LEVEL:       Debug level (0-5)
      LCMAPS_MOD_DIR:           Location of the LCMAPS library      
      LCMAPS_DIR:               Location of the LCMAPS config files 
      LCMAPS_ETC_DIR:           Location of the LCMAPS config files (=LCMAPS_DIR)
      LCMAPS_POLICY_STRING      The list of policies (default = NULL)
      LCMAPS_POLICY_ACQ_STRING  The list of acquisition policies (default = acquisition_policy)
      LCMAPS_POLICY_ENF_STRING  The list of enforcement policies (default = enforcement_policy)
**************************************************************************************************/
int setup_scas_config_name_value (config_name_values cnv, int n_config_name_values)
{
    int i = 0;


    for (i = 0; i < n_config_name_values; i++)
    {
        /* SCAS */
        if ((strcasecmp (cnv[i].name, "scas_port") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_PORT", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "scas_log_file") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_LOG_FILE", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "scas_log_type") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_LOG_TYPE", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "scas_log_string") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_LOG_STRING", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "scas_log_level") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_LOG_LEVEL", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "scas_debug_level") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_DEBUG_LEVEL", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "scas_hostcert") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_HOSTCERT", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "scas_hostkey") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_HOSTKEY", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "scas_capath") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_CAPATH", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "scas_no_daemonize") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_NO_DAEMONIZE", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "scas_refresh_rate") == 0) && (cnv[i].value))
        {
            setenv ("SCAS_REFRESH_RATE", cnv[i].value, 1);
        }

        /* VOMS Stuff */
        else if ((strcasecmp (cnv[i].name, "vomsdir") == 0) && (cnv[i].value))
        {
            setenv ("VOMSDIR", cnv[i].value, 1);
        }

        /* LCAS */
        else if ((strcasecmp (cnv[i].name, "lcas_log_file") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_LOG_FILE", cnv[i].value, 1);
        }
        else if ((strcasecmp (cnv[i].name, "lcas_log_type") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_LOG_TYPE", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_log_string") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_LOG_STRING", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_db_file") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_DB_FILE", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_log_level") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_LOG_LEVEL", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_debug_level") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_DEBUG_LEVEL", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_mod_dir") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_MOD_DIR", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_dir") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_DIR", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcas_etc_dir") == 0) && (cnv[i].value))
        {
            setenv ("LCAS_ETC_DIR", cnv[i].value, 1);
        } 

        /* LCMAPS */
        else if ((strcasecmp (cnv[i].name, "lcmaps_mapdir") == 0) && (cnv[i].value))
        {
            setenv ("MAPDIR", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_log_file") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_LOG_FILE", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_log_type") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_LOG_TYPE", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_log_string") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_LOG_STRING", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_db_file") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_DB_FILE", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_log_level") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_LOG_LEVEL", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_debug_level") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_DEBUG_LEVEL", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_mod_dir") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_MOD_DIR", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_dir") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_DIR", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_etc_dir") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_ETC_DIR", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_policy_string") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_POLICY_STRING", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_policy_acq_string") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_policy_acq_string", cnv[i].value, 1);
        } 
        else if ((strcasecmp (cnv[i].name, "lcmaps_policy_enf_string") == 0) && (cnv[i].value))
        {
            setenv ("LCMAPS_POLICY_ENF_STRING", cnv[i].value, 1);
        }

        /* Unsupported option */  
        else
        {
            fprintf (stderr, "SCAS: Configuration option '%s' not supported.\n", cnv[i].name);
            scas_log (1,           "Configuration option '%s' not supported.\n", cnv[i].name);
            return 1;
        }
    }

    return 0;
}


/**************************************************************************************************
function :                    get_scas_config_value_by_name 
                              (config_name_values cnv, int n_config_name_values)
description :
                   This function will setup all the SCAS, LCAS, and LCMAPS environment variables
      SCAS_PORT:              TCP portnumber for the SCAS service to listen on
      SCAS_LOG_FILE:          SCAS log file
      SCAS_HOSTCERT:          Host certificate of the SCAS service
      SCAS_HOSTKEY:           Private key file that's with the host certificate of the SCAS service
      SCAS_CAPATH:            CA certificate directory that contains the CRL files and the CA root certificates
      SCAS_LOG_LEVEL:         Log level (0-5)
      SCAS_DEBUG_LEVEL:       Debug level (0-5)

      LCAS_LOG_FILE:          Name and location of the logfile
      LCAS_LOG_TYPE:          Type of logging (logfile, syslog, both or none)
      LCAS_LOG_STRING:        Log string to be prepended to the logging lines
      LCAS_LOG_LEVEL:         Log level (0-5)
      LCAS_DB_FILE:           Name of the lcas policy file
      LCAS_DEBUG_LEVEL:       Debug level (0-5)
      LCAS_MOD_DIR:           Location of the LCAS library
      LCAS_DIR:               Location of the LCAS config files
      LCAS_ETC_DIR:           Location of the LCAS config files (=LCAS_DIR)

      LCMAPS_LOG_FILE:          Name and location of the logfile
      LCMAPS_LOG_TYPE:          Type of logging (logfile, syslog, both or none)
      LCMAPS_LOG_STRING:        Log string to be prepended to the logging lines
      LCMAPS_DB_FILE:           Name of the lcmaps policy file
      LCMAPS_LOG_LEVEL:         Log level (0-5)
      LCMAPS_DEBUG_LEVEL:       Debug level (0-5)
      LCMAPS_MOD_DIR:           Location of the LCMAPS library      
      LCMAPS_DIR:               Location of the LCMAPS config files 
      LCMAPS_ETC_DIR:           Location of the LCMAPS config files (=LCMAPS_DIR)
      LCMAPS_POLICY_STRING      The list of policies (default = NULL)
      LCMAPS_POLICY_ACQ_STRING  The list of acquisition policies (default = acquisition_policy)
      LCMAPS_POLICY_ENF_STRING  The list of enforcement policies (default = enforcement_policy)
**************************************************************************************************/
char * get_scas_config_value_by_name (config_name_values cnv, int n_config_name_values, char * search_name)
{
    int i = 0;


    for (i = 0; i < n_config_name_values; i++)
    {
        /* SCAS */
        if ((strcasecmp (cnv[i].name, search_name) == 0))
        {
            return (cnv[i].value);
        }
    }
    scas_log (1,           "Configuration option '%s' not configured.\n", search_name);
    return NULL;
}


/*******************************************************************************************
Function:                  print_config_name_value
Description:               Print the config_name_value structure to stdout
Parameters:
    config_name_values   cnv   :  the structure that will hold the name value pairs 
    int   n_config_name_values :  number of elements in the cnv structure
Return:
                   == 0 :  Success
                   != 0 :  Error has occured
*******************************************************************************************/
int print_config_name_value (config_name_values cnv, int n_config_name_values)
{
    int i = 0;

    for (i = 0; i < n_config_name_values; i++)
    {
        scas_log_debug (1, "Configuration name/value [%d]: name '%s', value '%s')\n", i, cnv[i].name, cnv[i].value);
    }

    return 0;
}



/*******************************************************************************************
Function:                  free_config_name_value
Description:               Will free the allocated memory for this array of structures and it struct contents
Parameters:
    config_name_values * cnv   :  the pointer to the structure that will hold the name value pairs 
    int   n_config_name_values :  number of elements in the cnv structure
Return:
                   == 0 :  Success
                   != 0 :  Error has occured
*******************************************************************************************/
int free_config_name_value (config_name_values * cnv, int n_config_name_values)
{
    int i = 0;

    for (i = 0; i < n_config_name_values; i++) 
    {
        free((*cnv)[i].name); 
        free((*cnv)[i].value); 
    }

    free(*cnv);
    *cnv = NULL;

    return 0;
}


/*******************************************************************************************
Function:                  open_scas_config_file
Description:               Open SCAS configuration file and return the name/value pairs with the amount of entries found.
                           In this situation I'm using the secure_open function, then I read it in the buffer and close it.
                           The configuration file may not be larger then BUFFER_SIZE_SCAS_CONFIG_FILE
Parameters:
    char * config_file_path    :  Path to the configuration file
    config_name_values * cnv   :  pointer to the structure that will hold the name value pairs 
    int * n_config_name_values :  number of elements in the cnv structure
Return:
                   == 0 :  Success
                   != 0 :  Error has occured
*******************************************************************************************/
int open_scas_config_file (char * config_file_path, config_name_values * cnv, int * n_config_name_values)
{
    unsigned char * buf = NULL;
    int             num_bytes_read = 0;
    int             fd = -1;


    if (!(config_file_path))
    {
        scas_log (1, "No config file provided\n");
        return 1;
    }
    else
    {
        scas_log (3, "Configuration file path to use is: %s\n", config_file_path);
    }
    

    /* 
     * Using secure opening of the configuration file 
     * File must be: 
     *     regular file
     *     0644
     *     Action is read only
     *     Explicit check on not being a symlink is enabled
     *     The file must be owned by the user who owns the process (the test should exclude sticky bit situations on the file and executable
     *     The file will not be checked on its group ownership. In this case, not a problem, because only the owner of the process could change the file. 
     */
    fd = secure_fileopen (config_file_path, (S_IFREG|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH), O_RDONLY,
                          TRUE, TRUE, FALSE);
    if (fd >= 0)
    {
        scas_log_debug (5, "Configuration file opened.\n");

        /* Allocate buffer */
        buf = (unsigned char *) malloc (sizeof(char) * BUFFER_SIZE_SCAS_CONFIG_FILE);
        if (!buf)
        {
            fprintf  (stderr, "SCAS: memory allocation failure, not enough memory available\n");
            scas_log (1,            "memory allocation failure, not enough memory available\n");
            return 1;
        }

        /* Read file content to buffer */
        num_bytes_read = readfdtobuf (fd, &buf, BUFFER_SIZE_SCAS_CONFIG_FILE);
        close (fd);
   
        if ( (num_bytes_read > 0) && (buf) )
        {
            scas_log_debug (5, "Configuration file loading into the buffer.\n");
            if (parse_scas_config_file (buf, num_bytes_read, cnv, n_config_name_values))
            {
                fprintf (stderr, "SCAS: Configuration file parsing failed.\n");
                scas_log (1,           "Configuration file parsing failed.\n");
                free(buf);
                return 1;
            }
            else
            {
                scas_log_debug (5, "Successfully parsed config file.\n");
                free(buf);
            }  
        }
        else
        {
            scas_log (1, "Failure: No bytes read from configuration file.\n");
            free(buf);
        }
    }
    else
    {
        fprintf (stderr, "SCAS: No file descripter returned\n");
        scas_log (1,           "No file descripter returned\n");
        return 1;
    }

    return 0;
}



/*******************************************************************************************
Function:                  secure_fileopen
Description:               use this function to be sure that this file can't be hijacked
                           before you read or write to it.
                           Various permission checks will be performed.
Parameters:
    const char * filepath : The full path to the file to open.
    int modeAndPerms      : Specify the specific file type, permissions to the file and in writing 
                            situation, it will specify the file permission on writing.
                            file modes:  S_IFREG, S_IFLNK, and others
                                         S_IRUSR, S_IWUSR, S_IXUSR, 
                                         S_IRGRP, S_IWGRP, S_IXGRP, 
                                         S_IROTH, S_IWOTH, S_IXOTH
                                         S_ISUID, S_ISGID, S_ISVTX                                     
                            Example:     (S_IFREG|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)
    int openflags         : Specifies for which purpose the file is going to be opened, read, write or both
                            Example:     O_RDONLY
Return:
                   => 0 :  Open File descriptor
                    < 0 :  Error code

*******************************************************************************************/
int secure_fileopen (char * filepath, int modeAndPerms, int openflags, 
                     int isNoSymLink, 
                     int realUIDmustOwnFile, 
                     int realGIDmustOwnFile)
{
    int         fd = 0;
    int         rc = 0;
    struct stat stat_before;
    char *      errorbuffer = malloc (sizeof (char) * MAX_LOG_BUFFER_SIZE);


    /* Error buffer */
    if (!errorbuffer)
    {
        scas_log (1, "Out of memory error, couldn't allocate %d bytes\n", MAX_LOG_BUFFER_SIZE);
        return -5;
    } 


    /* 
     * If the errmsg isn't set to a buffer or variable, then fail.
     * If the errmsg is set and it's pointer value is set to NULL, then no buffer is allocated. Allocate the buffer.
     * If the errmsg is set and not pointing to NULL, then assume that there is a buffer allocated
     */


    /* Input sanity */
    if ((!filepath) || (strlen(filepath) == 0))
    {
        scas_log (1, "filepath not supplied");
        rc = -10;
        goto end; 
    }


    /* Get the file status... */
    if (stat(filepath, &stat_before)) 
    {
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        scas_log (1, "stat on '%s' ended! %s\n", filepath, errorbuffer);
        rc = -15;
        goto end; 
    }

    /*  Check to see if the source proxy is indeed owned by the user that launched this program. */
    if (realUIDmustOwnFile)
    { 
        if (stat_before.st_uid != getuid()) 
        {
            strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
            scas_log (1, "The file '%s' is not owned by the calling user. %s\n", filepath, errorbuffer);
            rc = -20;
            goto end; 
        }
    }

    /*  Check to see if the source proxy is indeed owned by the group of the user that launched this program. */
    if (realGIDmustOwnFile)
    { 
        if (stat_before.st_uid != getgid()) 
        {
            strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
            scas_log (1, "The file '%s' is not owned by the group of the calling user. %s\n", filepath, errorbuffer);
            rc = -21;
            goto end; 
        }
    }


    /* Check to see if the source proxy is not a link. */
    if (isNoSymLink)
    { 
        if (S_ISLNK(stat_before.st_mode)) 
        {
            strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
            scas_log (1, "The file '%s' is a link and therefore rejected. %s\n", filepath, errorbuffer);
            rc = -22;
            goto end; 
        }
    }


    /* Check the permissions of the source proxy. */
    if (stat_before.st_mode != (stat_before.st_mode & modeAndPerms)) 
    {
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        fprintf (stderr, "SCAS: The file '%s' has wrong protection bits set. %s\n", filepath, errorbuffer);
        scas_log (1,           "The file '%s' has wrong protection bits set. %s\n", filepath, errorbuffer);
        rc = -23;
        goto end; 
    }


    /* Perform safe_open() */
    if ((fd = safe_open(filepath, openflags, modeAndPerms)) < 0)
    { 
        strerror_r (errno, errorbuffer, MAX_LOG_BUFFER_SIZE);
        scas_log (1, "safe_open() ended in a failure: %s\n", errorbuffer);
        rc = fd;
        goto end; 
    }


    rc = fd;

end:
    /* Pass the File descriptor in both good and error conditions */
    free(errorbuffer);
    return rc;
}


